@@ -18,6 +18,7 @@ except ImportError:
     sys.exit(1)
 
 from numpy import zeros, ones, log10, hstack, array, floor, mean, where, rint, inf
+import numpy as np
 
 # The peak decay rates (magic goes here :).
 PEAK_DECAY_RATE = 1.0 - 3E-6
@@ -25,10 +26,10 @@ PEAK_DECAY_RATE = 1.0 - 3E-6
 PEAK_FALLOFF_COUNT = 32 # default : 16
 
 class GLPlotWidget(QtGui.QWidget):
-    def __init__(self, parent=None, logger=None):
+    def __init__(self, parent, sharedGLWidget, logger=None):
         super(GLPlotWidget, self).__init__()
 
-        self.glWidget = GLWidget(self)
+        self.glWidget = GLWidget(self, sharedGLWidget)
         
         self.xmin = 0.1
         self.xmax = 1.
@@ -235,15 +236,26 @@ class GLPlotWidget(QtGui.QWidget):
         
         return x1, x2, n
 
-    def tree_rebin(self, y, ns):
-        i = 0
-        y2 = array([])        
+    def tree_rebin(self, y, ns, N):
+        y2 = zeros(N)
+
+        n = 0
         for i in range(len(ns)-1):
             y3 = y[ns[i]:ns[i+1]]
-            y3.shape = (len(y3)/2**i, 2**i)
-            y3 = mean(y3, axis=1)
+            d = 2**i
+            l = len(y3)/d
+            y3.shape = (l, d)
+
+            # Note: the FFT spectrum is mostly used to identify frequency content
+            # ans _peaks_ are particularly interesting (e.g. feedback frequencies)
+            # so we display the _max_ instead of the mean of each bin 
+            #y3 = mean(y3, axis=1)
             #y3 = (y3[::2] + y3[1::2])*0.5
-            y2 = hstack((y2, y3))
+            
+            y3 = np.max(y3, axis=1)
+
+            y2[n:n+len(y3)] = y3
+            n += l
         
         return y2
 
@@ -256,6 +268,10 @@ class GLPlotWidget(QtGui.QWidget):
             if self.logx:
                 self.transformed_x1, self.transformed_x2, n = self.pre_tree_rebin(x1, x2)
                 self.n = [0] + n
+                self.N = 0
+                for i in range(len(self.n)-1):
+                    self.N += (self.n[i+1] - self.n[i])/2**i
+
             else:
                 self.transformed_x1 = x1
                 self.transformed_x2 = x2
@@ -277,7 +293,7 @@ class GLPlotWidget(QtGui.QWidget):
         x2 = self.transformed_x2
         
         if self.logx:
-            y = self.tree_rebin(self.y, self.n)
+            y = self.tree_rebin(self.y, self.n, self.N)
         else:
             n = floor(1./(x2[2] - x1[1]))
             if n>0:
@@ -311,12 +327,33 @@ class GLPlotWidget(QtGui.QWidget):
         if self.peaks_enabled:
             transformed_peak = self.ytransform(self.peak)        
         
-            x1_with_peaks = hstack((x1, x1))
-            x2_with_peaks = hstack((x2, x2))
-            y_with_peaks = hstack((transformed_peak, transformed_y))
-            r_with_peaks = hstack((1.*Ones, 0.*Ones))
-            g_with_peaks = hstack((1. - self.peak_int, 0.5*Ones_shaded))
-            b_with_peaks = hstack((1. - self.peak_int, 0.*Ones))
+            n = x1.size
+
+            # FIXME should be done conditionally to need_transform
+            x1_with_peaks = zeros((2*n))
+            x2_with_peaks = zeros((2*n))
+            y_with_peaks = zeros((2*n))
+            r_with_peaks = zeros((2*n))
+            g_with_peaks = zeros((2*n))
+            b_with_peaks = zeros((2*n))
+
+            x1_with_peaks[:n] = x1
+            x1_with_peaks[n:] = x1
+
+            x2_with_peaks[:n] = x2
+            x2_with_peaks[n:] = x2
+
+            y_with_peaks[:n] = transformed_peak
+            y_with_peaks[n:] = transformed_y
+
+            r_with_peaks[:n] = 1.*Ones
+            r_with_peaks[n:] = 0.*Ones
+
+            g_with_peaks[:n] = 1. - self.peak_int
+            g_with_peaks[n:] = 0.5*Ones_shaded
+
+            b_with_peaks[:n] = 1. - self.peak_int
+            b_with_peaks[n:] = 0.*Ones
         else:
             x1_with_peaks = x1
             x2_with_peaks = x2
@@ -398,10 +435,8 @@ class GLPlotWidget(QtGui.QWidget):
 
 
 class GLWidget(QtOpenGL.QGLWidget):
-    def __init__(self, parent):
-        super(GLWidget, self).__init__(parent)
-        
-        self.parent = parent
+    def __init__(self, parent, sharedGLWidget):
+        super(GLWidget, self).__init__(parent, shareWidget=sharedGLWidget)
 
         self.lastPos = QtCore.QPoint()
         
@@ -428,6 +463,8 @@ class GLWidget(QtOpenGL.QGLWidget):
         # when QPainter.begin() is called.
         self.setAutoFillBackground(False)
 
+        self.gridList = None
+
     def minimumSizeHint(self):
         return QtCore.QSize(50, 50)
 
@@ -461,6 +498,47 @@ class GLWidget(QtOpenGL.QGLWidget):
         self.yMajorTick = yMajorTick
         self.yMinorTick = yMinorTick
 
+        self.updateGrid()
+
+    def updateGrid(self):
+        if self.gridList == None or self.gridList == 0:
+            return
+
+        w = self.width()
+        h = self.height()
+        
+        GL.glNewList(self.gridList, GL.GL_COMPILE)
+
+        self.qglColor(QtGui.QColor(Qt.Qt.gray))
+        for x in self.xMajorTick:        
+            GL.glBegin(GL.GL_LINES)
+            GL.glVertex2f(x, 0)
+            GL.glVertex2f(x, h)
+            GL.glEnd()
+        
+        self.qglColor(QtGui.QColor(Qt.Qt.lightGray))
+        for x in self.xMinorTick:        
+            GL.glBegin(GL.GL_LINES)
+            GL.glVertex2f(x, 0)
+            GL.glVertex2f(x, h)
+            GL.glEnd()
+            
+        self.qglColor(QtGui.QColor(Qt.Qt.gray))
+        for y in self.yMajorTick:        
+            GL.glBegin(GL.GL_LINES)
+            GL.glVertex2f(0, y)
+            GL.glVertex2f(w, y)
+            GL.glEnd()
+        
+        #GL.glColor3f(0.5, 0.5, 0.5)
+        #for y in self.yMinorTick:        
+        #    GL.glBegin(GL.GL_LINES)
+        #    GL.glVertex2f(0, y)
+        #    GL.glVertex2f(w, y)
+        #    GL.glEnd() 
+
+        GL.glEndList()
+
     #def paintGL(self):
     def paintEvent(self, event):
         self.makeCurrent()
@@ -575,8 +653,8 @@ class GLWidget(QtOpenGL.QGLWidget):
         if self.ruler:
             painter.setRenderHint(QtGui.QPainter.Antialiasing)
             
-            x = self.parent.inverseXTransform(self.mousex)
-            y = self.parent.inverseYTransform(self.height() - self.mousey)
+            x = self.parent().inverseXTransform(self.mousex)
+            y = self.parent().inverseYTransform(self.height() - self.mousey)
             text = "%d Hz, %.1f dB" %(x, y)
             
             # compute tracker bounding rect
@@ -606,6 +684,7 @@ class GLWidget(QtOpenGL.QGLWidget):
 
     def resizeGL(self, width, height):
         self.setupViewport(self.width(), self.height())
+        self.updateGrid()
 
     def setupViewport(self, width, height):
         GL.glViewport(0, 0, width, height)
@@ -628,36 +707,16 @@ class GLWidget(QtOpenGL.QGLWidget):
         GL.glEnd()
 
     def drawGrid(self):
-        w = self.width()
-        h = self.height()
-        
-        self.qglColor(QtGui.QColor(Qt.Qt.gray))
-        for x in self.xMajorTick:        
-            GL.glBegin(GL.GL_LINES)
-            GL.glVertex2f(x, 0)
-            GL.glVertex2f(x, h)
-            GL.glEnd()
-        
-        self.qglColor(QtGui.QColor(Qt.Qt.lightGray))
-        for x in self.xMinorTick:        
-            GL.glBegin(GL.GL_LINES)
-            GL.glVertex2f(x, 0)
-            GL.glVertex2f(x, h)
-            GL.glEnd()
-            
-        self.qglColor(QtGui.QColor(Qt.Qt.gray))
-        for y in self.yMajorTick:        
-            GL.glBegin(GL.GL_LINES)
-            GL.glVertex2f(0, y)
-            GL.glVertex2f(w, y)
-            GL.glEnd()
-        
-        #GL.glColor3f(0.5, 0.5, 0.5)
-        #for y in self.yMinorTick:        
-        #    GL.glBegin(GL.GL_LINES)
-        #    GL.glVertex2f(0, y)
-        #    GL.glVertex2f(w, y)
-        #    GL.glEnd() 
+        if self.gridList == None:
+            # display list used for the grid
+            self.gridList = GL.glGenLists(1)
+
+            if self.gridList == 0 or self.gridList == None:
+                raise RuntimeError( """Unable to generate a new display-list, context may not support display lists""")
+
+            self.updateGrid()
+
+        GL.glCallList(self.gridList)
 
     def drawBorder(self):
         w = self.width()
